library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity acceleration_block is
	port(
			clk		: in  std_logic;
			reset_n		: in  std_logic;
			read		: in  std_logic;
			write		: in  std_logic;
			chipselect	: in  std_logic;
			address		: in  std_logic_vector(4 downto 0);
			readdata	: out std_logic_vector(7 downto 0);
			writedata	: in  std_logic_vector(7 downto 0)
	);
end acceleration_block;

architecture implementation of acceleration_block is

  signal cpu_addr_m     : std_logic;
  signal cpu_ram_wren   : std_logic;
  signal cpu_ram_rden   : std_logic;
  signal reg_m_en       : std_logic;
  signal reg_m_reset    : std_logic;
  signal reg_m_start    : std_logic;
  signal reg_m          : std_logic_vector(2 downto 0);

  signal ram_wren_a      : std_logic;
  signal ram_addr_a      : std_logic_vector(4 downto 0);
  signal ram_from_data_a : std_logic_vector(6 downto 0);
  signal ram_to_data_a   : std_logic_vector(6 downto 0);
  signal ram_wren_b      : std_logic;
  signal ram_addr_b      : std_logic_vector(4 downto 0);
  signal ram_from_data_b : std_logic_vector(6 downto 0);
  signal ram_to_data_b   : std_logic_vector(6 downto 0);

  signal domove_addr_gen_a : std_logic_vector(4 downto 0);
  signal domove_addr_gen_b : std_logic_vector(4 downto 0);

  signal perm_addr_gen     : std_logic_vector(4 downto 0);
  
  signal fsm_mux_sel     : std_logic;
  signal fsm_reg_en      : std_logic;
  signal fsm_ram_wren_a  : std_logic;
  signal fsm_ram_wren_b  : std_logic;
  signal fsm_cpu_en      : std_logic;
  signal fsm_cnt_en      : std_logic;
  signal	fsm_cnt_reset   : std_logic;
  signal	fsm_perm_sel    : std_logic;
  signal	fsm_perm_wren   : std_logic;
  signal	fsm_perm_addr   : std_logic_vector(1 downto 0);
  signal fsm_step        : std_logic_vector(2 downto 0);	

  signal twist_data_a : std_logic_vector(6 downto 0);
  signal twist_data_b : std_logic_vector(6 downto 0);

  signal mux_data_a : std_logic_vector(6 downto 0);
  signal mux_data_b : std_logic_vector(6 downto 0);

  signal reg_data_a : std_logic_vector(6 downto 0);
  signal reg_data_b : std_logic_vector(6 downto 0);
  
  signal binary_cnt_int : integer range 0 to 31;
  signal binary_cnt     : std_logic_vector(4 downto 0);
  
  signal perm_addr      : std_logic_vector(4 downto 0);
  signal permtonum_out  : std_logic_vector(4 downto 0);

begin

  cpu_addr_m <= address(4) and not address(3) and address(2) and not address(1) and not address(0);
  cpu_ram_wren <= chipselect and write and not read and reset_n;
  cpu_ram_rden <= chipselect and not write and read and reset_n;

  with fsm_cpu_en select reg_m_en <=
    reg_m_reset                   when '0',
    cpu_addr_m and cpu_ram_wren   when others;

  reg_m_reset <= fsm_step(2) and fsm_step(1) and fsm_step(0);
  reg_m_start <= reg_m(2) and reg_m(1) and reg_m(0);

  process (clk)
  begin
    if rising_edge(clk) then
      if reset_n = '0' then
        reg_m <= "111";
      elsif reg_m_en = '1' then
        reg_m <= writedata(2 downto 0);
        if reg_m_reset = '1' then
          reg_m <= "111";
        end if;
      end if;
    end if;
  end process;

  ram_wren_a <= fsm_ram_wren_a;
  
  with fsm_perm_wren select perm_addr <=
    binary_cnt        when '0',
    perm_addr_gen     when others;
  
  with fsm_perm_sel select ram_addr_a <=
    domove_addr_gen_a when '0',
    perm_addr         when others;
  
  with fsm_perm_sel select ram_to_data_a <=
    mux_data_a             when '0',
    "00" & permtonum_out   when others;

  with fsm_cpu_en select ram_wren_b <=
    fsm_ram_wren_b when '0',
    cpu_ram_wren   when others;

  with fsm_cpu_en and reg_m_start select ram_addr_b <=
    domove_addr_gen_b when '0',
    address           when others;
  
  with fsm_cpu_en select ram_to_data_b <=
    mux_data_b            when '0',
    writedata(6 downto 0) when others;

  with cpu_ram_rden select readdata <=
    '0' & ram_from_data_b when '1',
    "11111111"            when others;

  process (clk)
  begin
    if rising_edge(clk) then
      if fsm_reg_en = '1' then
        reg_data_a <= twist_data_a;
        reg_data_b <= twist_data_b;
      end if;
    end if;
  end process;
    
  with fsm_mux_sel select mux_data_a <=
    twist_data_a when '0',
    reg_data_a   when others;

  with fsm_mux_sel select mux_data_b <=
    twist_data_b when '0',
    reg_data_b   when others;

  DUAL_PORT_RAM: entity work.true_dual_port_ram_single_clock port map(
	  
      clock => clk,
	
      address_a => ram_addr_a,
      data_a    => ram_to_data_a,
      wren_a    => ram_wren_a,
      q_a       => ram_from_data_a,

      address_b => ram_addr_b,
      data_b    => ram_to_data_b,
      wren_b    => ram_wren_b,
      q_b       => ram_from_data_b
  );

  FSM: entity work.moore_state_machine port map(
	  
     clk          => clk,
     reset_n      => reset_n,
     m            => reg_m,
	  cnt          => binary_cnt,
	  cnt_en       => fsm_cnt_en,
	  cnt_reset    => fsm_cnt_reset,
	  perm_sel     => fsm_perm_sel,
	  perm_wren    => fsm_perm_wren,
	  perm_addr    => fsm_perm_addr,
     mux_sel      => fsm_mux_sel,
     reg_en       => fsm_reg_en,
     ram_wren_a   => fsm_ram_wren_a,
     ram_wren_b   => fsm_ram_wren_b,
     cpu_en       => fsm_cpu_en,
     step         => fsm_step

  );

  TWIST: entity work.twist port map(

     m          => reg_m,
     step       => fsm_step,
     data_a_in  => ram_from_data_a,
     data_b_in  => ram_from_data_b,
     data_a_out => twist_data_a,
     data_b_out => twist_data_b

  );

  DOMOVE_ADDRESS_GENERATOR: entity work.domove_addr_gen port map(
	  
     m          => reg_m,
     step       => fsm_step,
     addr_a     => domove_addr_gen_a,
     addr_b     => domove_addr_gen_b

  );
  
  PERM_ADDRESS_GENERATOR: entity work.perm_addr_gen port map(
	  
     addr_sel   => binary_cnt(4 downto 2),
	  addr_out   => perm_addr_gen

  );

  binary_cnt <= std_logic_vector(to_unsigned(binary_cnt_int, 5));
  
  BINARY_COUNTER: entity work.binary_counter port map(
	  
      clk        => clk,
		reset      => fsm_cnt_reset,
		enable     => fsm_cnt_en,
		q          => binary_cnt_int

  );
  
  PERMTONUM: entity work.permtonum port map(
  
  		clk        => clk,
		data       => ram_from_data_a(6 downto 2),
		addr       => fsm_perm_addr,
		q          => permtonum_out
  );
  
end implementation;
