library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity de2_vga_raster is
  
  port (
    clk	       : in  std_logic;
    reset_n    : in  std_logic;
    write      : in  std_logic;
    chipselect : in  std_logic;
    address    : in  std_logic_vector(7 downto 0);
    writedata  : in  std_logic_vector(7 downto 0);

    VGA_CLK,                                 -- Clock
    VGA_HS,                                  -- H_SYNC
    VGA_VS,                                  -- V_SYNC
    VGA_BLANK,                               -- BLANK
    VGA_SYNC : out std_logic;                -- SYNC
    VGA_R,                                   -- Red[9:0]
    VGA_G,                                   -- Green[9:0]
    VGA_B : out std_logic_vector(9 downto 0) -- Blue[9:0]
    );
end de2_vga_raster;

architecture rtl of de2_vga_raster is

  -- Video parameters
  
  constant HTOTAL       : integer := 800;
  constant HSYNC        : integer := 96;
  constant HBACK_PORCH  : integer := 48;
  constant HACTIVE      : integer := 640;
  constant HFRONT_PORCH : integer := 16;
  
  constant VTOTAL       : integer := 525;
  constant VSYNC        : integer := 2;
  constant VBACK_PORCH  : integer := 33;
  constant VACTIVE      : integer := 480;
  constant VFRONT_PORCH : integer := 10;

  constant CUBIE_HEIGHT   : integer := 36;
  constant V1RegionOffset : integer := 60;
  constant V2RegionOffset : integer := 259;
  constant V3RegionOffset : integer := 176;
  constant V4RegionOffset : integer := 144;
  constant V5RegionOffset : integer := 160;
  constant V6RegionOffset : integer := 479;

  -- Signals for the video controller
  signal Hcount : unsigned(9 downto 0);  -- Horizontal position (0-800)
  signal Vcount : unsigned(9 downto 0);  -- Vertical position (0-524)
  signal MyHcount : unsigned(9 downto 0);
  signal MyVcount : unsigned(9 downto 0);
  signal EndOfLine, EndOfField : std_logic;
  signal vga_hblank, vga_hsync, vga_vblank, vga_vsync : std_logic;  -- Sync. signals
  signal clk25 : std_logic := '0';

  signal HRegion            : std_logic_vector(11 downto 0);
  signal V1Region, V2Region : std_logic_vector(6  downto 0);
  signal V3Region, V4Region : std_logic_vector(3  downto 0);
  signal V5Region, V6Region : std_logic_vector(3  downto 0);
  signal face               : std_logic_vector(2  downto 0);
  signal position           : std_logic_vector(3  downto 0);
  signal OneHot             : std_logic_vector(63 downto 0);
  
  signal SBorderHRegion                   : std_logic_vector(7 downto 0);
  signal SBorderV1Region, SBorderV2Region : std_logic_vector(3 downto 0);
  signal SBorderV3Region, SBorderV4Region : std_logic_vector(1 downto 0);
  signal SBorderV5Region, SBorderV6Region : std_logic_vector(1 downto 0);
  signal SBorder                          : std_logic_vector(5 downto 0);

  signal ram_address    : unsigned(5 downto 0);
  signal raddr      	: integer range 0 to 63;
  signal raddr_reg1 	: integer range 0 to 63;
  signal raddr_reg2 	: integer range 0 to 63;
  signal lines      	: std_logic;
  signal color      	: std_logic_vector(2 downto 0);
  signal border_reg 	: std_logic_vector(5 downto 0);
  signal Face4line, Face4hline1, Face4hline2, Face4hline4              : std_logic;
  signal            Face4vline1, Face4vline2, Face4vline3, Face4vline4 : std_logic;
  signal Face5line, Face5hline1, Face5hline2, Face5hline4              : std_logic;
  signal            Face5vline1, Face5vline2, Face5vline3, Face5vline4 : std_logic;
  signal Face6line, Face6hline1, Face6hline2, Face6hline3              : std_logic;
  signal            Face6vline1, Face6vline2                           : std_logic;

  signal raddr_text      	: unsigned(7 downto 0);
  signal text_addr      	: std_logic_vector(5 downto 0);
  signal text_data      	: std_logic_vector(7 downto 0);
  signal ram_selection, text_display, text_region1       : std_logic;
  signal text_region23_v, text_region2_h, text_region3_h : std_logic;
  signal subtraction_result     : std_logic_vector(6 downto 0);
  signal text_region            : std_logic;

begin

  ram_address <= unsigned(address(5 downto 0)); 
  ram_selection <= address(7) or address(6);

  RAM_COLOR: entity work.simple_dual_port_ram_single_clock_color port map(
     clk          => clk,
     raddr        => raddr,
     waddr        => to_integer(ram_address),
     data         => writedata(2 downto 0),
     we           => chipselect and write and not ram_selection,
     q            => color
  );

  RAM_TEXT: entity work.simple_dual_port_ram_single_clock_text port map(
     clk          => clk,
     raddr        => to_integer(raddr_text),
     waddr        => to_integer(unsigned(address(7 downto 0))) - 64,
     data         => writedata(5 downto 0),
     we           => chipselect and write and ram_selection,
     q            => text_addr
  );

  CHAR_ROM: entity work.char_rom port map (
    clock         => clk,     
    address       => text_addr(5 downto 0) & std_logic_vector(MyVcount(2 downto 0)),
    q             => text_data
  );

  --25M clk
  
  process (clk)
  begin
    if rising_edge(clk) then
      clk25 <= not clk25;
    end if;
  end process;

  -- Horizontal and vertical counters

  HCounter : process (clk25)
  begin
    if rising_edge(clk25) then      
      if reset_n = '0' then
        Hcount <= (others => '0');
      elsif EndOfLine = '1' then
        Hcount <= (others => '0');
      else
        Hcount <= Hcount + 1;
      end if;      
    end if;
  end process HCounter;

  EndOfLine <= '1' when Hcount = HTOTAL - 1 else '0';
  
  VCounter: process (clk25)
  begin
    if rising_edge(clk25) then      
      if reset_n = '0' then
        Vcount <= (others => '0');
      elsif EndOfLine = '1' then
        if EndOfField = '1' then
          Vcount <= (others => '0');
        else
          Vcount <= Vcount + 1;
        end if;
      end if;
    end if;
  end process VCounter;

  EndOfField <= '1' when Vcount = VTOTAL - 1 else '0';

  -- State machines to generate HSYNC, VSYNC, HBLANK, and VBLANK

  HSyncGen : process (clk25)
  begin
    if rising_edge(clk25) then     
      if reset_n = '0' or EndOfLine = '1' then
        vga_hsync <= '1';
      elsif Hcount = HSYNC - 1 then
        vga_hsync <= '0';
      end if;
    end if;
  end process HSyncGen;
  
  HBlankGen : process (clk25)
  begin
    if rising_edge(clk25) then
      if reset_n = '0' then
        vga_hblank <= '1';
      elsif Hcount = HSYNC + HBACK_PORCH then
        vga_hblank <= '0';
      elsif Hcount = HSYNC + HBACK_PORCH + HACTIVE then
        vga_hblank <= '1';
      end if;      
    end if;
  end process HBlankGen;

  VSyncGen : process (clk25)
  begin
    if rising_edge(clk25) then
      if reset_n = '0' then
        vga_vsync <= '1';
      elsif EndOfLine ='1' then
        if EndOfField = '1' then
          vga_vsync <= '1';
        elsif Vcount = VSYNC - 1 then
          vga_vsync <= '0';
        end if;
      end if;      
    end if;
  end process VSyncGen;

  VBlankGen : process (clk25)
  begin
    if rising_edge(clk25) then    
      if reset_n = '0' then
        vga_vblank <= '1';
      elsif EndOfLine = '1' then
        if Vcount = VSYNC + VBACK_PORCH - 1 then
          vga_vblank <= '0';
        elsif Vcount = VSYNC + VBACK_PORCH + VACTIVE - 1 then
          vga_vblank <= '1';
        end if;
      end if;
    end if;
  end process VBlankGen;

  -- Cube generator

  MyHcount <= Hcount - HSYNC - HBACK_PORCH;
  MyVcount <= Vcount - VSYNC - VBACK_PORCH;
  
  HRegion(11) <= '1' when MyHcount(9 downto 5) = 7  else '0';
  HRegion(10) <= '1' when MyHcount(9 downto 5) = 8  else '0';
  HRegion(9)  <= '1' when MyHcount(9 downto 5) = 9  else '0';
  HRegion(8)  <= '1' when MyHcount(9 downto 5) = 10 else '0';
  HRegion(7)  <= '1' when MyHcount(9 downto 5) = 11 else '0';
  HRegion(6)  <= '1' when MyHcount(9 downto 5) = 12 else '0';
  HRegion(5)  <= '1' when MyHcount(9 downto 5) = 3  else '0';
  HRegion(4)  <= '1' when MyHcount(9 downto 5) = 4  else '0';
  HRegion(3)  <= '1' when MyHcount(9 downto 5) = 5  else '0';
  HRegion(2)  <= '1' when MyHcount(9 downto 5) = 14 else '0';
  HRegion(1)  <= '1' when MyHcount(9 downto 5) = 15 else '0';
  HRegion(0)  <= '1' when MyHcount(9 downto 5) = 16 else '0';

  V1Region(0) <= '1' when MyVcount < MyHcount/2                  - V1RegionOffset + 12 else '0';
  V1Region(1) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT   - V1RegionOffset + 8  else '0';
  V1Region(2) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*2 - V1RegionOffset + 4  else '0';
  V1Region(3) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*3 - V1RegionOffset      else '0';
  V1Region(4) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*4 - V1RegionOffset      else '0';
  V1Region(5) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*5 - V1RegionOffset      else '0';
  V1Region(6) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*6 - V1RegionOffset      else '0';

  V2Region(0) <= '1' when MyVcount < V2RegionOffset - MyHcount/2                  + 12 else '0';
  V2Region(1) <= '1' when MyVcount < V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT   + 8  else '0';
  V2Region(2) <= '1' when MyVcount < V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*2 + 4  else '0';
  V2Region(3) <= '1' when MyVcount < V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*3      else '0';
  V2Region(4) <= '1' when MyVcount < V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*4      else '0';
  V2Region(5) <= '1' when MyVcount < V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*5      else '0';
  V2Region(6) <= '1' when MyVcount < V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*6      else '0';

  V3Region(0) <= '1' when MyVcount < MyHcount/2                  - V3RegionOffset      else '0';
  V3Region(1) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT   - V3RegionOffset      else '0';
  V3Region(2) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*2 - V3RegionOffset      else '0';
  V3Region(3) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*3 - V3RegionOffset      else '0';

  V4Region(0) <= '1' when MyVcount < V4RegionOffset - MyHcount/2                       else '0';
  V4Region(1) <= '1' when MyVcount < V4RegionOffset - MyHcount/2 + CUBIE_HEIGHT        else '0';
  V4Region(2) <= '1' when MyVcount < V4RegionOffset - MyHcount/2 + CUBIE_HEIGHT*2      else '0';
  V4Region(3) <= '1' when MyVcount < V4RegionOffset - MyHcount/2 + CUBIE_HEIGHT*3      else '0';

  V5Region(0) <= '1' when MyVcount < MyHcount/2                  + V5RegionOffset + 12 else '0';
  V5Region(1) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT   + V5RegionOffset + 8  else '0';
  V5Region(2) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*2 + V5RegionOffset + 4  else '0';
  V5Region(3) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*3 + V5RegionOffset      else '0';

  V6Region(0) <= '1' when MyVcount < V6RegionOffset - MyHcount/2                  + 12 else '0';
  V6Region(1) <= '1' when MyVcount < V6RegionOffset - MyHcount/2 + CUBIE_HEIGHT   + 8  else '0';
  V6Region(2) <= '1' when MyVcount < V6RegionOffset - MyHcount/2 + CUBIE_HEIGHT*2 + 4  else '0';
  V6Region(3) <= '1' when MyVcount < V6RegionOffset - MyHcount/2 + CUBIE_HEIGHT*3      else '0';

  FaceDetector : process (clk25)
  begin
    if rising_edge(clk25) then
      if((HRegion(11) or HRegion(10) or HRegion(9)) and (V1Region(6) and not V1Region(3))) = '1' then
        face <= "000";
      elsif((HRegion(8) or HRegion(7) or HRegion(6)) and (V2Region(6) and not V2Region(3))) = '1' then
        face <= "001";
      elsif((V1Region(3) and not V1Region(0)) and (V2Region(3) and not V2Region(0))) = '1' then
        face <= "010";
      elsif((HRegion(0) or HRegion(1) or HRegion(2)) and (V3Region(3) and not V3Region(0))) = '1' then
        face <= "011";
      elsif((HRegion(3) or HRegion(4) or HRegion(5)) and (V4Region(3) and not V4Region(0))) = '1' then
        face <= "100";
      elsif((V5Region(3) and not V5Region(0)) and (V6Region(3) and not V6Region(0))) = '1' then
        face <= "101";
      else
        face <= "111";
      end if;
    end if;
  end process FaceDetector;

  PositionDetector : process (clk25)
  begin
    if rising_edge(clk25) then
      if(OneHot(0) or OneHot(9) or OneHot(18) or OneHot(27) or OneHot(36) or OneHot(45)) = '1' then
        position <= "0000";
      elsif(OneHot(1) or OneHot(10) or OneHot(19) or OneHot(28) or OneHot(37) or OneHot(46)) = '1' then
        position <= "0001";
      elsif(OneHot(2) or OneHot(11) or OneHot(20) or OneHot(29) or OneHot(38) or OneHot(47)) = '1' then
        position <= "0010";
      elsif(OneHot(3) or OneHot(12) or OneHot(21) or OneHot(30) or OneHot(39) or OneHot(48)) = '1' then
        position <= "0011";
      elsif(OneHot(4) or OneHot(13) or OneHot(22) or OneHot(31) or OneHot(40) or OneHot(49)) = '1' then
        position <= "0100";
      elsif(OneHot(5) or OneHot(14) or OneHot(23) or OneHot(32) or OneHot(41) or OneHot(50)) = '1' then
        position <= "0101";
      elsif(OneHot(6) or OneHot(15) or OneHot(24) or OneHot(33) or OneHot(42) or OneHot(51)) = '1' then
        position <= "0110";
      elsif(OneHot(7) or OneHot(16) or OneHot(25) or OneHot(34) or OneHot(43) or OneHot(52)) = '1' then
        position <= "0111";
      elsif(OneHot(8) or OneHot(17) or OneHot(26) or OneHot(35) or OneHot(44) or OneHot(53)) = '1' then
        position <= "1000";
      else
        position <= "0000"; 
      end if;
    end if;
  end process PositionDetector;

  raddr <= to_integer(unsigned(face(2 downto 0) & "000") + unsigned("000" & face(2 downto 0))
                    + unsigned("00" & position(3 downto 0)));

  OneHot(0)  <= HRegion(11) and V1Region(4) and not V1Region(3); --face1
  OneHot(1)  <= HRegion(10) and V1Region(4) and not V1Region(3);
  OneHot(2)  <= HRegion(9)  and V1Region(4) and not V1Region(3);
  OneHot(3)  <= HRegion(11) and V1Region(5) and not V1Region(4);
  OneHot(4)  <= HRegion(10) and V1Region(5) and not V1Region(4);
  OneHot(5)  <= HRegion(9)  and V1Region(5) and not V1Region(4);
  OneHot(6)  <= HRegion(11) and V1Region(6) and not V1Region(5);
  OneHot(7)  <= HRegion(10) and V1Region(6) and not V1Region(5);
  OneHot(8)  <= HRegion(9)  and V1Region(6) and not V1Region(5);

  OneHot(9)  <= HRegion(8)  and V2Region(4) and not V2Region(3); --face2
  OneHot(10) <= HRegion(7)  and V2Region(4) and not V2Region(3);
  OneHot(11) <= HRegion(6)  and V2Region(4) and not V2Region(3);
  OneHot(12) <= HRegion(8)  and V2Region(5) and not V2Region(4);
  OneHot(13) <= HRegion(7)  and V2Region(5) and not V2Region(4);
  OneHot(14) <= HRegion(6)  and V2Region(5) and not V2Region(4);
  OneHot(15) <= HRegion(8)  and V2Region(6) and not V2Region(5);
  OneHot(16) <= HRegion(7)  and V2Region(6) and not V2Region(5);
  OneHot(17) <= HRegion(6)  and V2Region(6) and not V2Region(5);

  OneHot(18) <= V1Region(1) and not V1Region(0) and V2Region(1) and not V2Region(0); --face3
  OneHot(19) <= V1Region(1) and not V1Region(0) and V2Region(2) and not V2Region(1);
  OneHot(20) <= V1Region(1) and not V1Region(0) and V2Region(3) and not V2Region(2);
  OneHot(21) <= V1Region(2) and not V1Region(1) and V2Region(1) and not V2Region(0);
  OneHot(22) <= V1Region(2) and not V1Region(1) and V2Region(2) and not V2Region(1);
  OneHot(23) <= V1Region(2) and not V1Region(1) and V2Region(3) and not V2Region(2);
  OneHot(24) <= V1Region(3) and not V1Region(2) and V2Region(1) and not V2Region(0);
  OneHot(25) <= V1Region(3) and not V1Region(2) and V2Region(2) and not V2Region(1);
  OneHot(26) <= V1Region(3) and not V1Region(2) and V2Region(3) and not V2Region(2);

  OneHot(27) <= HRegion(0)  and V3Region(1) and not V3Region(0); --face4
  OneHot(28) <= HRegion(1)  and V3Region(1) and not V3Region(0);
  OneHot(29) <= HRegion(2)  and V3Region(1) and not V3Region(0);
  OneHot(30) <= HRegion(0)  and V3Region(2) and not V3Region(1);
  OneHot(31) <= HRegion(1)  and V3Region(2) and not V3Region(1);
  OneHot(32) <= HRegion(2)  and V3Region(2) and not V3Region(1);
  OneHot(33) <= HRegion(0)  and V3Region(3) and not V3Region(2);
  OneHot(34) <= HRegion(1)  and V3Region(3) and not V3Region(2);
  OneHot(35) <= HRegion(2)  and V3Region(3) and not V3Region(2);

  OneHot(36) <= HRegion(3)  and V4Region(1) and not V4Region(0); --face5
  OneHot(37) <= HRegion(4)  and V4Region(1) and not V4Region(0);
  OneHot(38) <= HRegion(5)  and V4Region(1) and not V4Region(0);
  OneHot(39) <= HRegion(3)  and V4Region(2) and not V4Region(1);
  OneHot(40) <= HRegion(4)  and V4Region(2) and not V4Region(1);
  OneHot(41) <= HRegion(5)  and V4Region(2) and not V4Region(1);
  OneHot(42) <= HRegion(3)  and V4Region(3) and not V4Region(2);
  OneHot(43) <= HRegion(4)  and V4Region(3) and not V4Region(2);
  OneHot(44) <= HRegion(5)  and V4Region(3) and not V4Region(2);

  OneHot(45) <= V5Region(3) and not V5Region(2) and V6Region(3) and not V6Region(2); --face6
  OneHot(46) <= V5Region(2) and not V5Region(1) and V6Region(3) and not V6Region(2);
  OneHot(47) <= V5Region(1) and not V5Region(0) and V6Region(3) and not V6Region(2);
  OneHot(48) <= V5Region(3) and not V5Region(2) and V6Region(2) and not V6Region(1);
  OneHot(49) <= V5Region(2) and not V5Region(1) and V6Region(2) and not V6Region(1);
  OneHot(50) <= V5Region(1) and not V5Region(0) and V6Region(2) and not V6Region(1);
  OneHot(51) <= V5Region(3) and not V5Region(2) and V6Region(1) and not V6Region(0);
  OneHot(52) <= V5Region(2) and not V5Region(1) and V6Region(1) and not V6Region(0);
  OneHot(53) <= V5Region(1) and not V5Region(0) and V6Region(1) and not V6Region(0);

  FacehlineGen : process (clk25)
  begin
    if rising_edge(clk25) then    
      if MyHcount(9 downto 5) = 10 then
        Face4hline1 <= '1';
      elsif MyHcount(9 downto 5) = 14 then
        Face4hline1 <= '0';
      end if;

      if MyHcount(9 downto 5) = 13 then
        Face4hline2 <= '1';
      elsif MyHcount(9 downto 5) = 17 then
        Face4hline2 <= '0';
      end if;

      if MyHcount(9 downto 5) = 13 then
        Face4hline4 <= '1';
      else
        Face4hline4 <= '0';
      end if;

      if MyHcount(9 downto 5) = 6 then
        Face5hline1 <= '1';
      elsif MyHcount(9 downto 5) = 10 then
        Face5hline1 <= '0';
      end if;

      if MyHcount(9 downto 5) = 3 then
        Face5hline2 <= '1';
      elsif MyHcount(9 downto 5) = 7 then
        Face5hline2 <= '0';
      end if;

      if MyHcount(9 downto 5) = 6 then
        Face5hline4 <= '1';
      else
        Face5hline4 <= '0';
      end if;

      if MyHcount = 224 or MyHcount = 225 then
        Face6hline1 <= '1';
      else
        Face6hline1 <= '0';
      end if;

      if MyHcount = 320 or MyHcount = 321 then
        Face6hline2 <= '1';
      else
        Face6hline2 <= '0';
      end if;

      if MyHcount = 416 or MyHcount = 417 then
        Face6hline3 <= '1';
      else
        Face6hline3 <= '0';
      end if;
    end if;
  end process FacehlineGen;

  Face4vline1 <= '1' when MyVcount = V2RegionOffset - MyHcount/2                  + 12 else '0';
  Face4vline2 <= '1' when MyVcount = V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*3      else '0';
  Face4vline3 <= '1' when MyVcount = V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*6      else '0';
  Face4vline4 <= '1' when MyVcount = V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*3 + 12 else '0';
  Face4line   <= (Face4hline1 and Face4vline1) or (Face4hline2 and Face4vline2) or (Face4hline2 and Face4vline3) or (Face4hline4 and Face4vline4);

  Face5vline1 <= '1' when MyVcount = MyHcount/2                  - V1RegionOffset + 12 else '0';
  Face5vline2 <= '1' when MyVcount = MyHcount/2 + CUBIE_HEIGHT*3 - V1RegionOffset      else '0';
  Face5vline3 <= '1' when MyVcount = MyHcount/2 + CUBIE_HEIGHT*6 - V1RegionOffset      else '0';
  Face5vline4 <= '1' when MyVcount = MyHcount/2 + CUBIE_HEIGHT*3 - V1RegionOffset + 12 else '0';
  Face5line   <= (Face5hline1 and Face5vline1) or (Face5hline2 and Face5vline2) or (Face5hline2 and Face5vline3) or (Face5hline4 and Face5vline4);

  Face6vline1 <= '1' when MyVcount > 266 and MyVcount < 380 else '0';
  Face6vline2 <= '1' when MyVcount > 314 and MyVcount < 428 else '0';
  Face6line   <= (Face6hline1 and Face6vline1) or (Face6hline2 and Face6vline2) or (Face6hline3 and Face6vline1);

  RaddrReg : process (clk25)
  begin
    if rising_edge(clk25) then    
      raddr_reg1 <= raddr;   
      raddr_reg2 <= raddr_reg1;
    end if;
  end process RaddrReg;

  lines <= '0' when raddr_reg1 = raddr and raddr_reg1 = raddr_reg2 and (Face4line or Face5line or Face6line) = '0' else '1';

  --Selection Border Register
  
  process (clk)
  begin
    if rising_edge(clk) then
      if reset_n = '0' then
        border_reg <= "000001";
      elsif (chipselect and write) = '1' and address = "00110110" then
        border_reg <= writedata(5 downto 0);
      end if;
    end if;
  end process;

  -- Selection border generator

  SBorderHRegion(7)  <= '1' when MyHcount(9 downto 3) > 10 else '0';
  SBorderHRegion(6)  <= '1' when MyHcount(9 downto 3) > 24 else '0';
  SBorderHRegion(5)  <= '1' when MyHcount(9 downto 3) > 54 else '0';
  SBorderHRegion(4)  <= '1' when MyHcount(9 downto 3) > 68 else '0';
  SBorderHRegion(3)  <= '1' when MyHcount(9 downto 3) > 26 else '0';
  SBorderHRegion(2)  <= '1' when MyHcount(9 downto 3) > 40 else '0';
  SBorderHRegion(1)  <= '1' when MyHcount(9 downto 3) > 38 else '0';
  SBorderHRegion(0)  <= '1' when MyHcount(9 downto 3) > 52 else '0';

  SBorderV1Region(0) <= '1' when MyVcount < MyHcount/2                  - V1RegionOffset + 4 else '0';
  SBorderV1Region(1) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*3 - V1RegionOffset - 8 else '0';
  SBorderV1Region(2) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*3 - V1RegionOffset + 8 else '0';
  SBorderV1Region(3) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*6 - V1RegionOffset + 8 else '0';

  SBorderV2Region(0) <= '1' when MyVcount < V2RegionOffset - MyHcount/2                  + 4 else '0';
  SBorderV2Region(1) <= '1' when MyVcount < V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*3 - 8 else '0';
  SBorderV2Region(2) <= '1' when MyVcount < V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*3 + 8 else '0';
  SBorderV2Region(3) <= '1' when MyVcount < V2RegionOffset - MyHcount/2 + CUBIE_HEIGHT*6 + 8 else '0';

  SBorderV3Region(0) <= '1' when MyVcount < MyHcount/2                  - V3RegionOffset - 8 else '0';
  SBorderV3Region(1) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*3 - V3RegionOffset + 8 else '0';

  SBorderV4Region(0) <= '1' when MyVcount < V4RegionOffset - MyHcount/2                  - 8 else '0';
  SBorderV4Region(1) <= '1' when MyVcount < V4RegionOffset - MyHcount/2 + CUBIE_HEIGHT*3 + 8 else '0';
  
  SBorderV5Region(0) <= '1' when MyVcount < MyHcount/2                  + V5RegionOffset + 4 else '0';
  SBorderV5Region(1) <= '1' when MyVcount < MyHcount/2 + CUBIE_HEIGHT*3 + V5RegionOffset + 8 else '0';

  SBorderV6Region(0) <= '1' when MyVcount < V6RegionOffset - MyHcount/2                  + 4 else '0';
  SBorderV6Region(1) <= '1' when MyVcount < V6RegionOffset - MyHcount/2 + CUBIE_HEIGHT*3 + 8 else '0';

  SBorderGenerator : process (clk25)
  begin
    if rising_edge(clk25) then
      SBorder(0) <= SBorderHRegion(3) and not SBorderHRegion(2) and SBorderV1Region(3) and not SBorderV1Region(1) and
                 not ((HRegion(11) or HRegion(10) or HRegion(9)) and V1Region(6) and not V1Region(3));
      SBorder(1) <= SBorderHRegion(1) and not SBorderHRegion(0) and SBorderV2Region(3) and not SBorderV2Region(1) and
                 not ((HRegion(8)  or HRegion(7)  or HRegion(6)) and V2Region(6) and not V2Region(3));
      SBorder(2) <= SBorderV1Region(2) and not SBorderV1Region(0) and SBorderV2Region(2) and not SBorderV2Region(0) and
                 not (V1Region(3) and not V1Region(0) and V2Region(3) and not V2Region(0));
      SBorder(3) <= SBorderHRegion(5) and not SBorderHRegion(4) and SBorderV3Region(1) and not SBorderV3Region(0) and
                 not ((HRegion(2)  or HRegion(1)  or HRegion(0)) and V3Region(3) and not V3Region(0));
      SBorder(4) <= SBorderHRegion(7) and not SBorderHRegion(6) and SBorderV4Region(1) and not SBorderV4Region(0) and
                 not ((HRegion(5)  or HRegion(4)  or HRegion(3)) and V4Region(3) and not V4Region(0));
      SBorder(5) <= SBorderV5Region(1) and not SBorderV5Region(0) and SBorderV6Region(1) and not SBorderV6Region(0) and
                 not (V5Region(3) and not V5Region(0) and V6Region(3) and not V6Region(0));
    end if;
  end process SBorderGenerator;

  -- Text Display
  
  Mux_8_to_1: entity work.mux_8_to_1 port map(
     data_in      => text_data(7 downto 0),
     sel          => std_logic_vector(MyHcount(2 downto 0)),
     data_out     => text_display
  );

  TextRegion: process (clk25)
  begin
    if rising_edge(clk25) then
      if MyVcount(9 downto 3) = 3 then
        if MyHcount(9 downto 3) = 25 then
          text_region1 <= '1';
        elsif MyHcount(9 downto 3) = 55 then
          text_region1 <= '0';
        end if;
      else
        text_region1 <= '0';
      end if;

      if MyVcount(9 downto 3) = 39 then
        text_region23_v <= '1';
      elsif MyVcount(9 downto 3) = 43 then
        text_region23_v <= '0';
      end if;

      if MyHcount(9 downto 3) = 4 then
        text_region2_h <= '1';
      elsif MyHcount(9 downto 3) = 24 then
        text_region2_h <= '0';
      end if;

      if MyHcount(9 downto 3) = 56 then
        text_region3_h <= '1';
      elsif MyHcount(9 downto 3) = 76 then
        text_region3_h <= '0';
      end if;

      text_region <= text_region1 or ((text_region2_h or text_region3_h) and text_region23_v);
    end if;
  end process TextRegion;

  subtraction_result <= std_logic_vector(MyVcount(9 downto 3)-39);

  raddr_text  <= 
    unsigned('0' & std_logic_vector(MyHcount(9 downto 3)))-25 when text_region1 = '1' else
    26 + unsigned(subtraction_result(3 downto 0) & "0000") + unsigned(subtraction_result(5 downto 0) & "00")
       + unsigned('0' & std_logic_vector(MyHcount(9 downto 3)))
      when (text_region23_v and text_region2_h) = '1' else
    54 + unsigned(subtraction_result(3 downto 0) & "0000") + unsigned(subtraction_result(5 downto 0) & "00")
       + unsigned('0' & std_logic_vector(MyHcount(9 downto 3)))
      when (text_region23_v and text_region3_h) = '1';

  -- Registered video signals going to the video DAC

  VideoOut: process (clk25, reset_n)
  begin
    if reset_n = '0' then
      VGA_R <= "0000000000";
      VGA_G <= "0000000000";
      VGA_B <= "0000000000";
    elsif rising_edge(clk25) then
      if lines = '1' or (text_region and text_display) = '1' then
	VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";
      elsif ((border_reg(0) and SBorder(0)) or
             (border_reg(1) and SBorder(1)) or
             (border_reg(2) and SBorder(2)) or
             (border_reg(3) and SBorder(3)) or
             (border_reg(4) and SBorder(4)) or
             (border_reg(5) and SBorder(5))) = '1' then
        VGA_R <= "1010110100";
	VGA_G <= "1110000100";
	VGA_B <= "1110000100";
      elsif color = "110" then --yellow
	VGA_R <= "1111111111";
        VGA_G <= "1111111111";
        VGA_B <= "0000000000";		  
      elsif color = "010" then --red
	VGA_R <= "1111111111";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";	  
      elsif color = "011" then --blue
	VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "1111111111";		
      elsif color = "101" then --green
	VGA_R <= "0000000000";
        VGA_G <= "1111111111";
        VGA_B <= "0000000000";
      elsif color = "001" then --white
	VGA_R <= "1111111111";
        VGA_G <= "1111111111";
        VGA_B <= "1111111111";
      elsif color = "100" then --orange
	VGA_R <= "1111111111";
        VGA_G <= "0110001000";
        VGA_B <= "0001000000";
      elsif vga_hblank = '0' and vga_vblank = '0' then
        VGA_R <= "1111111111";
        VGA_G <= "1111111111";
        VGA_B <= "1111111111";
      else
        VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";    
      end if;
    end if;
  end process VideoOut;

  VGA_CLK <= clk25;
  VGA_HS <= not vga_hsync;
  VGA_VS <= not vga_vsync;
  VGA_SYNC <= '0';
  VGA_BLANK <= not (vga_hsync or vga_vsync);

end rtl;
